#include <stdio.h>
#include <windows.h>
#include <setupapi.h>
#include <devguid.h>
#include <regstr.h>

#include "cyioctl.h"
#include "cyusb30_def.h"
#include "mystuff.h"


#define vERBOSE
#define EP_OUT          0x01
#define EP_IN           0x81
#define MAXFX3BULKOUT   16384  // DMA_OUT buffersize in bytes
//#define MAXFX3BULKOUT   64  // DMA_OUT buffersize in bytes
#define MAXFX3BULKIN    1024
#define BLOCKSIZE       (64*MAXFX3BULKIN) // byte
#define TIMEOUT_I       1000
#define TIMEOUT_O       1000

enum e_cmd {e_cmd_read=1,e_cmd_write=2,e_cmd_readzlp=5,e_cmds};

#ifdef NOTNOW
#endif /* NOTNOW */



/* {AE18AA60-7F6A-11d4-97DD-00010229B959} */
static GUID CYUSBDRV_GUID = {0xae18aa60, 0x7f6a, 0x11d4, {0x97, 0xdd, 0x0, 0x1, 0x2, 0x29, 0xb9, 0x59}};

#define BYTES_PER_LINE 16
void display_le(FILE *f,unsigned char *d,unsigned int l,unsigned int addr_off) {
	int i,j;
	for(i=BYTES_PER_LINE;i;i--) fprintf(f," %02x",i-1);
	fprintf(f,"\n");
	for(i=0;i<l;i+=BYTES_PER_LINE){
		for(j=i+BYTES_PER_LINE;j>l;j--) fprintf(f,"   ");
		for(;j>i;j--) fprintf(f," %02x",d[j-1]);
		fprintf(f," : %08x\n",addr_off+i);
	}
}

LPSTR ErrorAsString(DWORD errorMessageID)
{
	if(errorMessageID == 0) return NULL; //No error message has been recorded
	LPSTR messageBuffer = NULL;
	FormatMessageA(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
			NULL, errorMessageID, MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), (LPSTR)&messageBuffer, 0, NULL);
	fprintf(stderr,"%s",messageBuffer);
	//Free the buffer.
	//LocalFree(messageBuffer);
	return messageBuffer;
}

DWORD in(HANDLE hUsb,unsigned char* data, int bufsize /* bytes */, DWORD *ptransferred) {
DWORD r;
unsigned char XmitBuf[sizeof(SINGLE_TRANSFER)];
OVERLAPPED ovLapStatus;
DWORD LastError;

memset(XmitBuf,0,sizeof(SINGLE_TRANSFER));
       ((PSINGLE_TRANSFER)XmitBuf)->ucEndpointAddress = EP_IN;
       ((PSINGLE_TRANSFER)XmitBuf)->IsoPacketLength   = 0;
       ((PSINGLE_TRANSFER)XmitBuf)->BufferOffset      = 0;
       ((PSINGLE_TRANSFER)XmitBuf)->BufferLength      = 0;

memset(&ovLapStatus,0,sizeof(OVERLAPPED));
   if(!(ovLapStatus.hEvent=CreateEvent(NULL, 0, 0, NULL))) {
		fprintf(stderr,"From CreateEvent...\n");
		LocalFree(ErrorAsString(GetLastError()));
		return(-1);
	}


// for bufsize == 0 (from request for ZLP) we adjust bufsize to 1 to avoid errors
        if (!bufsize) bufsize++;
	DeviceIoControl (hUsb,
			IOCTL_ADAPT_SEND_NON_EP0_DIRECT,
			&XmitBuf,
			sizeof(SINGLE_TRANSFER),
			data,
			bufsize,
			ptransferred,
			&ovLapStatus);

	LastError  = GetLastError();
	if(LastError == ERROR_SUCCESS);              // good
	else if(LastError == ERROR_IO_PENDING) {     // alright
#ifdef VERBOSE
		fprintf(stderr,"DeviceIoControl:\n");
		LocalFree(ErrorAsString(GetLastError()));
#endif /* VERBOSE */
	}
	else {                                       // bad
		fprintf(stderr,"DeviceIoControl:\n");
		LocalFree(ErrorAsString(GetLastError()));
		return (-1);
	}

	if(LastError == ERROR_IO_PENDING) { 
		r=WaitForSingleObject(ovLapStatus.hEvent,TIMEOUT_I); // ms
		if(r!=WAIT_OBJECT_0) {
			fprintf(stderr,"WaitForSingleObject:\n");
			LocalFree(ErrorAsString(GetLastError()));
			if(LastError == ERROR_IO_PENDING) {    // time out
				fprintf(stderr,"Timeout after %d ms.\n",TIMEOUT_I);
				return (-1);
			}
			else return (-1);
		}
	}

	if(!GetOverlappedResult(hUsb, &ovLapStatus, ptransferred, FALSE)){
		fprintf(stderr,"GetOverlappedResult:\n");
		LocalFree(ErrorAsString(GetLastError()));
		return(-1);
	}

	if(!CloseHandle(ovLapStatus.hEvent)) {
		fprintf(stderr,"CloseHandle:\n");
		LocalFree(ErrorAsString(GetLastError()));
		return(-1);
	}
	return (0);
}

DWORD out(HANDLE hUsb,unsigned char* data, int bufsize /* bytes */, DWORD *ptransferred) {
DWORD r;
unsigned char XmitBuf[sizeof(SINGLE_TRANSFER)];
OVERLAPPED ovLapStatus;
DWORD LastError;

      memset(XmitBuf,0,sizeof(SINGLE_TRANSFER));
      ((PSINGLE_TRANSFER)XmitBuf)->ucEndpointAddress = EP_OUT;
      ((PSINGLE_TRANSFER)XmitBuf)->IsoPacketLength   = 0;
      ((PSINGLE_TRANSFER)XmitBuf)->BufferOffset      = 0;
      ((PSINGLE_TRANSFER)XmitBuf)->BufferLength      = 0;

      memset(&ovLapStatus,0,sizeof(OVERLAPPED));
      if(!(ovLapStatus.hEvent=CreateEvent(NULL, 0, 0, NULL))) {
         fprintf(stderr,"From CreateEvent...\n");
         LocalFree(ErrorAsString(GetLastError()));
         return(-1);
         }

      DeviceIoControl (hUsb,
                       IOCTL_ADAPT_SEND_NON_EP0_DIRECT,
                       &XmitBuf,
                       sizeof(SINGLE_TRANSFER),
                       data,
                       bufsize,
                       ptransferred,
                       &ovLapStatus);

      LastError  = GetLastError();
      if(LastError == ERROR_SUCCESS);              // good
      else if(LastError == ERROR_IO_PENDING) {     // alright
#ifdef VERBOSE
         fprintf(stderr,"DeviceIoControl:\n");
         LocalFree(ErrorAsString(GetLastError()));
#endif /* VERBOSE */
	 }
      else {                                       // bad
         fprintf(stderr,"DeviceIoControl:\n");
         LocalFree(ErrorAsString(GetLastError()));
         return(-1);
         }
      
      if(LastError == ERROR_IO_PENDING) { 
         r=WaitForSingleObject(ovLapStatus.hEvent,TIMEOUT_O); // ms
         if(r!=WAIT_OBJECT_0) {
            fprintf(stderr,"WaitForSingleObject:\n");
            LocalFree(ErrorAsString(GetLastError()));
            return(-1);
            }
         }

      if(!GetOverlappedResult(hUsb, &ovLapStatus, ptransferred, FALSE)){
         fprintf(stderr,"GetOverlappedResult:\n");
         LocalFree(ErrorAsString(GetLastError()));
         return(-1);
	 }

      if(!CloseHandle(ovLapStatus.hEvent)) {
         fprintf(stderr,"CloseHandle:\n");
         LocalFree(ErrorAsString(GetLastError()));
         return(-1);
         }
return (0);
}

int main( int argc, char *argv[ ], char *envp[ ] )
{
HDEVINFO hDevInfo;
SP_DEVINFO_DATA devInfoData;
SP_DEVICE_INTERFACE_DATA  devInterfaceData;
PSP_INTERFACE_DEVICE_DETAIL_DATA functionClassDeviceData = NULL;
//SP_INTERFACE_DEVICE_DETAIL_DATA tmpInterfaceDeviceDetailData;
ULONG requiredLength = 0;
//char DevPath[USB_STRING_MAXLEN];
unsigned char buf[BLOCKSIZE+sizeof(SINGLE_TRANSFER)];
unsigned char* data=buf+sizeof(SINGLE_TRANSFER);
DWORD transferred=0;
ULONG length;
HANDLE hUsb,hFile;
int transfers,somesize,restsize /* bytes */ ,bufsize /* bytes */,j;
int someaddress /* byte address */;
char somestr[200];
SET_TRANSFER_SIZE_INFO SetTransferInfo;


fprintf(stdout,"HiHo\n");

// SetLastError(25);
// LocalFree(ErrorAsString(GetLastError()));

// Open a handle to the plug and play dev node.
// SetupDiGetClassDevs() returns a device information set that contains info on all
// installed devices of a specified class which are present.

hDevInfo = SetupDiGetClassDevs(&CYUSBDRV_GUID,
                               0, // Enumerator
                               0, // hwndParent
			       DIGCF_PRESENT|DIGCF_INTERFACEDEVICE);
//                               DIGCF_PRESENT | DIGCF_ALLCLASSES );
       
if (hDevInfo == INVALID_HANDLE_VALUE) {
   fprintf(stderr,"SetupDiGetClassDevs:\n");
   LocalFree(ErrorAsString(GetLastError()));
   goto fin;
   }

// If we have more than one device we need to iterate
// For now (and probably forever) we are going for the first
devInterfaceData.cbSize = sizeof(devInterfaceData);
if(!SetupDiEnumDeviceInterfaces ( hDevInfo, 
			          0, // DeviceInfoData
				  &CYUSBDRV_GUID, 
				  0, // Number in list, we take first if any
				  &devInterfaceData)) {
   fprintf(stderr,"SetupDiEnumdeviceInterfaces:\n");
   LocalFree(ErrorAsString(GetLastError()));
   fprintf(stderr,"...which probably means there are no devices with a CYUSB driver powered on.\n");
   goto fin;
  }	

// Allocate a function class device data structure to receive the goods about this 
// particular device.
//if(!
		SetupDiGetInterfaceDeviceDetail ( hDevInfo, // Device Info Set
			              &devInterfaceData, 
				      NULL, // DeviceInterfaceDetailData
				      0,    // DeviceInterfaceDetailDataSize
                                      &requiredLength,
				      NULL)//) { // DeviceInfoData
//   LocalFree(ErrorAsString(GetLastError()));
//   fprintf(stderr,"...from SetupDiGetInterfaceDeviceDetail\n");
//  fprintf(stderr, "required length %ld\n",requiredLength);
//   goto fin;
//  }	
;
ULONG predictedLength = requiredLength;

functionClassDeviceData = (PSP_INTERFACE_DEVICE_DETAIL_DATA) malloc (predictedLength);
functionClassDeviceData->cbSize =  sizeof (SP_INTERFACE_DEVICE_DETAIL_DATA);
devInfoData.cbSize = sizeof(devInfoData);
//Retrieve the information from Plug and Play */
if (!SetupDiGetInterfaceDeviceDetail (hDevInfo,
                    &devInterfaceData,
                    functionClassDeviceData,
                    predictedLength,
                    &requiredLength,
                    &devInfoData)) { 
   fprintf(stderr,"SetupDiGetInterfaceDeviceDetail:\n");
   LocalFree(ErrorAsString(GetLastError()));
   fprintf(stderr, "required length %ld\n",requiredLength);
   goto fin;
  }	
/* 
 * NOTE : x64 packing issue ,requiredLength return 5byte size of the (SP_INTERFACE_DEVICE_DETAIL_DATA) 
 * and functionClassDeviceData needed sizeof functionClassDeviceData 8byte 
*/										   
// int pathLen = requiredLength - functionClassDeviceData->cbSize;

//int pathLen = requiredLength - 
//	      (sizeof (tmpInterfaceDeviceDetailData.cbSize)+
//	      sizeof (tmpInterfaceDeviceDetailData.DevicePath));


//memcpy (DevPath, functionClassDeviceData->DevicePath, pathLen);
//DevPath[pathLen] = 0;
//fprintf(stdout,"Dev Path %s\n",DevPath);
fprintf(stdout,"functionClassDeviceData->DevicePath (not a joke):\n%s\n",
        functionClassDeviceData->DevicePath);

hUsb = CreateFile (functionClassDeviceData->DevicePath,
                    GENERIC_WRITE | GENERIC_READ,
                    FILE_SHARE_WRITE | FILE_SHARE_READ,
                    NULL,
                    OPEN_EXISTING,
                    FILE_FLAG_OVERLAPPED,
                    NULL);

if(!hUsb) {
   fprintf(stderr,"CreateFile:\n");
   LocalFree(ErrorAsString(GetLastError()));
   goto fin;
   }	
//Sleep(INFINITE);

/*******
 * We have a handle: hUsb
 * Action starts here
*******/
//fprintf(stdout,"We finally got hUsb=%d\n",(int)hUsb);

/*** Get transfer size EP_OUT from cyusb driver ***/

SetTransferInfo.EndpointAddress = EP_OUT;
SetTransferInfo.TransferSize    = 0;
fprintf(stderr,"DeviceIoControl:\n");
if(!DeviceIoControl(hUsb,
                    IOCTL_ADAPT_GET_TRANSFER_SIZE,
                    &SetTransferInfo, sizeof(SET_TRANSFER_SIZE_INFO),
                    &SetTransferInfo, sizeof(SET_TRANSFER_SIZE_INFO),
                    &transferred,
                    NULL)) {
   LocalFree(ErrorAsString(GetLastError()));
   goto fin;
   }
if(transferred) {
   fprintf(stdout,"EP: 0x%02x, transfer size = %lu\n",EP_OUT,SetTransferInfo.TransferSize);
   }
else {
   fprintf(stdout,"EP: 0x%02x, GET_TRANSFER_SIZE returned %lu bytes\n",EP_OUT,transferred);
   }

/*** Get transfer size EP_IN from cyusb driver ***/

SetTransferInfo.EndpointAddress = EP_IN;
SetTransferInfo.TransferSize    = 0;
fprintf(stderr,"DeviceIoControl:\n");
if(!DeviceIoControl(hUsb,
                    IOCTL_ADAPT_GET_TRANSFER_SIZE,
                    &SetTransferInfo, sizeof(SET_TRANSFER_SIZE_INFO),
                    &SetTransferInfo, sizeof(SET_TRANSFER_SIZE_INFO),
                    &transferred,
                    NULL)) {
   LocalFree(ErrorAsString(GetLastError()));
   goto fin;
   }
if(transferred) {
   fprintf(stdout,"EP: 0x%02x, transfer size = %lu\n",EP_IN,SetTransferInfo.TransferSize);
   }
else {
   fprintf(stdout,"EP: 0x%02x, GET_TRANSFER_SIZE returned %lu bytes\n",EP_IN,transferred);
   }

/*** User class request ***/

fprintf(stdout,"User class request GET_FW_VERSION %02x ... ",GET_FW_VERSION);
length = sizeof(SINGLE_TRANSFER) + 32; // 32 bytes of data expected
((PSINGLE_TRANSFER)buf)->SetupPacket.bmReqType.Direction = BMREQUEST_DEVICE_TO_HOST;
((PSINGLE_TRANSFER)buf)->SetupPacket.bmReqType.Type      = BMREQUEST_CLASS;
((PSINGLE_TRANSFER)buf)->SetupPacket.bmReqType.Recipient = BMREQUEST_TO_DEVICE;
((PSINGLE_TRANSFER)buf)->SetupPacket.bRequest            = GET_FW_VERSION;
((PSINGLE_TRANSFER)buf)->SetupPacket.wVal.hiByte         = 0;
((PSINGLE_TRANSFER)buf)->SetupPacket.wVal.lowByte        = 0;
((PSINGLE_TRANSFER)buf)->SetupPacket.wLength             = 32; // 32 bytes expected
((PSINGLE_TRANSFER)buf)->SetupPacket.ulTimeOut           = 5; // ms ?
((PSINGLE_TRANSFER)buf)->BufferLength = ((PSINGLE_TRANSFER)buf)->SetupPacket.wLength;
((PSINGLE_TRANSFER)buf)->BufferOffset = sizeof(SINGLE_TRANSFER);

if(!DeviceIoControl(hUsb,
                    IOCTL_ADAPT_SEND_EP0_CONTROL_TRANSFER,
                    buf,      // to device
                    length,
                    buf,      // from device
                    length,
                    &transferred,
                    NULL)) { // Overlapped
   fprintf(stderr,"DeviceIoControl:\n");
   LocalFree(ErrorAsString(GetLastError()));
   goto fin;
   }
fprintf(stdout,"returns %ld bytes: \n",transferred-sizeof(SINGLE_TRANSFER));
fprintf(stdout,"Firmware string: %s\n",data);

/*** User class request ***/

fprintf(stdout,"User class request GET_BUFF_SIZE %02x ... ",GET_BUFF_SIZE);
length = sizeof(SINGLE_TRANSFER) + 8; // 8 bytes of data expected
((PSINGLE_TRANSFER)buf)->SetupPacket.bmReqType.Direction = BMREQUEST_DEVICE_TO_HOST;
((PSINGLE_TRANSFER)buf)->SetupPacket.bmReqType.Type      = BMREQUEST_CLASS;
((PSINGLE_TRANSFER)buf)->SetupPacket.bmReqType.Recipient = BMREQUEST_TO_DEVICE;
((PSINGLE_TRANSFER)buf)->SetupPacket.bRequest            = GET_BUFF_SIZE;
((PSINGLE_TRANSFER)buf)->SetupPacket.wVal.hiByte         = 0;
((PSINGLE_TRANSFER)buf)->SetupPacket.wVal.lowByte        = 0;
((PSINGLE_TRANSFER)buf)->SetupPacket.wLength             = 8; // 8 bytes of data expected
((PSINGLE_TRANSFER)buf)->SetupPacket.ulTimeOut           = 5; // ms ?
((PSINGLE_TRANSFER)buf)->BufferLength = ((PSINGLE_TRANSFER)buf)->SetupPacket.wLength;
((PSINGLE_TRANSFER)buf)->BufferOffset = sizeof(SINGLE_TRANSFER);

if(!DeviceIoControl(hUsb,
                    IOCTL_ADAPT_SEND_EP0_CONTROL_TRANSFER,
                    buf,      // to device
                    length,
                    buf,      // from device
                    length,
                    &transferred,
                    NULL)) { // Overlapped
   fprintf(stderr,"DeviceIoControl:\n");
   LocalFree(ErrorAsString(GetLastError()));
   goto fin;
   }
fprintf(stdout,"returns %ld bytes: \n",transferred-sizeof(SINGLE_TRANSFER));
fprintf(stdout,"Host to FX3 (OUT), dma buffer size = 0x%04x, dma buffer count = 0x%04x\n",
               *((unsigned short *) (data+0)),*((unsigned short *) (data+2)));
fprintf(stdout,"FX3 to host (IN) , dma buffer size = 0x%04x, dma buffer count = 0x%04x\n",
               *((unsigned short *) (data+4)),*((unsigned short *) (data+6)));
for(transfers=1;;transfers++) {
   fprintf(stdout,"\n-------------------------------------\n");
   fprintf(stdout,"{i[n],o[ut],e[xit]} size (hex, 32-bit words) or\n");
   fprintf(stdout,"{re[ad],rz[lp],wr[ite]} address (hex, byte) size (hex, byte) or\n");
   fprintf(stdout,"{rf[ile],wf[ile]} address (hex, byte) size (hex, byte) filename>");
   if((fscanf(stdin,"%s",somestr) != 1)) { 
      fprintf(stderr,"read error\n");
      continue;
      }
   if(somestr[0]=='i' || somestr[0]=='I') { 
/***
 * in command   
***/
      if((fscanf(stdin,"%x",&somesize) != 1)) {
         fprintf(stderr,"read error\n");
         continue;
         }

      bufsize = somesize * sizeof(int);
      if(bufsize>BLOCKSIZE) {
         fprintf(stderr,"Warning: size (bytes) = %08x > %08x\n",bufsize,BLOCKSIZE);
         }
      if(in(hUsb, data, bufsize /* byte */, &transferred /* byte */)) {
         fprintf(stderr,"function in() failed\n"); 
         goto fin; 
         }
      fprintf(stdout,"transfer %d: 0x%08lx of 0x%08x bytes read\n",
              transfers,transferred,bufsize); 
      if(transferred) display_le(stdout,data,transferred,0);
      }
   else if(somestr[0]=='o' || somestr[0]=='O') {
/***
 * out command   
***/
      if((fscanf(stdin,"%x",&somesize) != 1)) {
         fprintf(stderr,"read error\n");
         continue;
         }
      bufsize = somesize * sizeof(int);
      if(bufsize>MAXFX3BULKOUT) {
         fprintf(stderr,"Warning: size (bytes) = %08x > %08x \n",bufsize,MAXFX3BULKOUT);
         }
//      fprintf(stdout,"buffer size will always be 0x%02x (bytes)\n",MAXFX3BULKOUT);
      for(j=0;
            j<somesize;
//          j<((somesize<MAXFX3BULKOUT/sizeof(int)) ?
//                      somesize : (MAXFX3BULKOUT/sizeof(int)));
          j++) {
         fprintf(stdout," word 0x%02x>",j);
         fscanf(stdin,"%x",((int *)data)+j);
         }
//      somesize=MAXFX3BULKOUT/(sizeof (int)); // force to EP out packet size
//      for(;j<somesize;j++) ((int *)data)[j]=0;
      if(somesize) display_le(stdout,data,somesize*sizeof(int),0);
      if(out(hUsb, data, bufsize /* byte */, &transferred /* byte */)) {
         fprintf(stderr,"function out() failed\n"); 
         goto fin; 
         }
      fprintf(stdout,"transfer %d: 0x%08lx of 0x%08x bytes written\n",
              transfers,transferred,somesize*sizeof(int)); 
      } 
   else if(!strncasecmp(somestr,"re",2)) {
/***
 * re command   
***/
      ((int *)data)[0]=e_cmd_read;
      if((fscanf(stdin,"%x",&someaddress) != 1)) {
         fprintf(stderr,"read error\n");
         continue;
         }
      ((int *)data)[1]=someaddress;
      if((fscanf(stdin,"%x",&bufsize) != 1)) {
         fprintf(stderr,"read error\n");
         continue;
         }
      if((bufsize/sizeof(int))*sizeof(int)<bufsize) bufsize=((bufsize/sizeof(int))+1)*sizeof(int);
      if(bufsize > BLOCKSIZE) {
         fprintf(stderr,"bufsize 0x%08x > max=0x%08x\n",bufsize,BLOCKSIZE);
         goto fin;
         }
      ((int *)data)[2]=bufsize;
// send out command
      if(out(hUsb, data, 3*sizeof(int) /* byte */, &transferred /* byte */)) {
         fprintf(stderr,"function out() failed\n"); 
         goto fin; 
         }
      transfers++;
// send in command
      if(in(hUsb, data, bufsize /* byte */, &transferred /* byte */)) {
         fprintf(stderr,"function in() failed\n"); 
         goto fin; 
         }
      fprintf(stdout,"0x%08lx of 0x%08x bytes read\n",transferred,bufsize); 
      if(transferred) display_le(stdout,data,transferred,someaddress);
      }
   else if(!strncasecmp(somestr,"rz",2)) {
/***
 * rz command   
***/
      ((int *)data)[0]=e_cmd_readzlp;
      if((fscanf(stdin,"%x",&someaddress) != 1)) {
         fprintf(stderr,"read error\n");
         continue;
         }
      ((int *)data)[1]=someaddress;
      if((fscanf(stdin,"%x",&bufsize) != 1)) {
         fprintf(stderr,"read error\n");
         continue;
         }
      if((bufsize/sizeof(int))*sizeof(int)<bufsize) bufsize=((bufsize/sizeof(int))+1)*sizeof(int);
      if(bufsize > BLOCKSIZE) {
         fprintf(stderr,"bufsize 0x%08x > max=0x%08x\n",bufsize,BLOCKSIZE);
         goto fin;
         }
      ((int *)data)[2]=bufsize;
// send out command
      if(out(hUsb, data, 3*sizeof(int) /* byte */, &transferred /* byte */)) {
         fprintf(stderr,"function out() failed\n"); 
         goto fin; 
         }
      transfers++;
// send in command
      if(in(hUsb, data, bufsize /* byte */, &transferred /* byte */)) {
         fprintf(stderr,"function in() failed\n"); 
         goto fin; 
         }
      fprintf(stdout,"0x%08lx of 0x%08x bytes read\n",transferred,bufsize); 
      if(transferred) display_le(stdout,data,transferred,someaddress);
      }
   else if(!strncasecmp(somestr,"wr",2)) {
/***
 * wr command   
***/
      ((int *)data)[0]=e_cmd_write;
      if((fscanf(stdin,"%x",&someaddress) != 1)) {
         fprintf(stderr,"read error\n");
         continue;
         }
      ((int *)data)[1]=someaddress;
      if((fscanf(stdin,"%x",&bufsize) != 1)) {
         fprintf(stderr,"read error\n");
         continue;
         }
      if((bufsize/sizeof(int))*sizeof(int)<bufsize) bufsize=((bufsize/sizeof(int))+1)*sizeof(int);
      if(bufsize > (BLOCKSIZE-3*sizeof(int))) {//  adjust for command words
         fprintf(stderr,"bufsize 0x%08x > max=0x%08x\n",bufsize,BLOCKSIZE-3*sizeof(int));
         goto fin;
         }
      ((int *)data)[2]=bufsize;
      for(j=3;
            j<(bufsize/sizeof(int)+3);
//          j<((somesize<MAXFX3BULKOUT/sizeof(int)) ?
//                      somesize : (MAXFX3BULKOUT/sizeof(int)));
         j++) {
         fprintf(stdout," word 0x%02x>",j-3);
         fscanf(stdin,"%x",((int *)data)+j);
         }
// send out command
      restsize =  bufsize+3*sizeof(int);
      while(restsize>MAXFX3BULKOUT) {
         if(out(hUsb, data+bufsize+3*sizeof(int)-restsize, MAXFX3BULKOUT /* byte */, &transferred /* byte */)) {
            fprintf(stderr,"function out() failed\n"); 
            goto fin; 
            }
         restsize -= transferred;
         }
      fprintf(stdout,"0x%08lx of 0x%08x bytes written (including 12 byte for command words)\n",
              transferred,bufsize+3*sizeof(int)); 
      if(out(hUsb, data+bufsize+3*sizeof(int)-restsize, restsize /* byte */, &transferred /* byte */)) {
         fprintf(stderr,"function out() failed\n"); 
         goto fin; 
            }
      fprintf(stdout,"0x%08lx of 0x%08x bytes written (including 12 byte for command words)\n",
              transferred,bufsize+3*sizeof(int)); 
//      if(transferred) 
      display_le(stdout,data+3*sizeof(int),bufsize,0);
      }
   else if(!strncasecmp(somestr,"rf",2)) {
/***
 * rf command: read from USB, write to file   
***/
      ((int *)data)[0]=e_cmd_read;
      if((fscanf(stdin,"%x",&someaddress) != 1)) {
         fprintf(stderr,"read error\n");
         continue;
         }
      ((int *)data)[1]=someaddress;
      if((fscanf(stdin,"%x",&bufsize) != 1)) {
         fprintf(stderr,"read error\n");
         continue;
         }
      if((bufsize/sizeof(int))*sizeof(int)<bufsize) bufsize=((bufsize/sizeof(int))+1)*sizeof(int);
      if(bufsize > BLOCKSIZE) {
         fprintf(stderr,"bufsize 0x%08x > max=0x%08x\n",bufsize,BLOCKSIZE);
         goto fin;
         }
      ((int *)data)[2]=bufsize;
      if((fscanf(stdin,"%s",somestr) != 1)) {
         fprintf(stderr,"read error\n");
         continue;
         }
      hFile = CreateFile(somestr,                // name of the write
                         GENERIC_WRITE,          // open for writing
                         0,                      // do not share
                         NULL,                   // default security
                         CREATE_NEW,             // create new file only
                         FILE_ATTRIBUTE_NORMAL,  // normal file
                         NULL);                  // no attr. template

      if (hFile == INVALID_HANDLE_VALUE) { 
         fprintf(stderr,"CreateFile %s:\n",somestr);
         LocalFree(ErrorAsString(GetLastError()));
         goto fin;
         }
// send out command
      if(out(hUsb, data, 3*sizeof(int) /* byte */, &transferred /* byte */)) {
         fprintf(stderr,"function out() failed\n"); 
         goto fin; 
         }
      transfers++;
// send in command
      if(in(hUsb, data, bufsize /* byte */, &transferred /* byte */)) {
         fprintf(stderr,"function in() failed\n"); 
         goto fin; 
         }
      fprintf(stdout,"0x%08lx of 0x%08x bytes read\n",transferred,bufsize); 
      somesize = transferred;
      for(restsize=transferred,transferred=0;restsize>0;restsize-=transferred){
         if(!WriteFile(hFile,            // file handle
                       data+(somesize-restsize), // start of data to write
                       restsize,         // number of bytes to write
                       &transferred,     // number of bytes that were written
                       NULL)){           // no overlapped structure
            fprintf(stderr,"WriteFile %s:\n",somestr);
            LocalFree(ErrorAsString(GetLastError()));
            goto fin;
            }
#ifdef VERBOSE
         fprintf(stdout,"Transferred %ld bytes to \"%s\"\n",transferred,somestr);
#endif /* VERBOSE */
         }
      fprintf(stdout,"Transfer of %d bytes to \"%s\" completed.\n",somesize,somestr);
      CloseHandle(hFile);
      }
   else if(!strncasecmp(somestr,"wf",2)) {
/***
 * wf command: write to USB, read from file     
***/
      ((int *)data)[0]=e_cmd_write;
      if((fscanf(stdin,"%x",&someaddress) != 1)) {
         fprintf(stderr,"read error\n");
         continue;
         }
      ((int *)data)[1]=someaddress;
      if((fscanf(stdin,"%x",&bufsize) != 1)) {
         fprintf(stderr,"read error\n");
         continue;
         }
      if((bufsize/sizeof(int))*sizeof(int)<bufsize) bufsize=((bufsize/sizeof(int))+1)*sizeof(int);
      if(bufsize > (BLOCKSIZE-3*sizeof(int))) {//  adjust for command words
         fprintf(stderr,"bufsize 0x%08x > max=0x%08x\n",bufsize,BLOCKSIZE-3*sizeof(int));
         goto fin;
         }
      ((int *)data)[2]=bufsize;
      if((fscanf(stdin,"%s",somestr) != 1)) {
         fprintf(stderr,"read error\n");
         continue;
         }
      hFile = CreateFile(somestr,                // name of the read
                         GENERIC_READ,           // open for read
                         0,                      // do not share
                         NULL,                   // default security
                         OPEN_EXISTING,          // open existing file only
                         FILE_ATTRIBUTE_NORMAL,  // normal file
                         NULL);                  // no attr. template

      if (hFile == INVALID_HANDLE_VALUE) { 
         fprintf(stderr,"CreateFile (existing) %s:\n",somestr);
         LocalFree(ErrorAsString(GetLastError()));
         goto fin;
         }
      somesize = GetFileSize(hFile,NULL);
      if (somesize == INVALID_FILE_SIZE) { 
         fprintf(stderr,"GetFileSize %s:\n",somestr);
         LocalFree(ErrorAsString(GetLastError()));
         goto fin;
         }
      if(somesize<bufsize) {
         fprintf(stderr,"\"%s\": only 0x%08x bytes\n",somestr,somesize);
         goto fin;
         }
      else if(somesize>bufsize) {
         fprintf(stderr,"WARNING \"%s\": 0x%08x bytes in file, only 0x%08x bytes requested.\n",
                 somestr,somesize,bufsize);
         }
      for(restsize = bufsize,transferred=0;restsize>0;restsize-=transferred){
         if(!ReadFile(hFile,                 // file handle
                       data+(3*sizeof(int))
                           +(bufsize-restsize),  // start of data to read
                       restsize,             // number of bytes to read
                       &transferred,         // number of bytes that were read
                       NULL)){               // no overlapped structure
            fprintf(stderr,"WriteFile %s:\n",somestr);
            LocalFree(ErrorAsString(GetLastError()));
            goto fin;
            }
#ifdef VERBOSE
         fprintf(stdout,"Transferred %ld bytes from \"%s\"\n",transferred,somestr);
#endif /* VERBOSE */
         }
      fprintf(stdout,"Transfer of %d bytes from \"%s\" completed.\n",somesize,somestr);
// send out command
      restsize =  bufsize+3*sizeof(int);
      while(restsize>MAXFX3BULKOUT) {
         if(out(hUsb, data+bufsize+3*sizeof(int)-restsize, MAXFX3BULKOUT /* byte */, &transferred /* byte */)) {
            fprintf(stderr,"function out() failed\n"); 
            goto fin; 
            }
         restsize -= transferred;
         }
      if(out(hUsb, data+bufsize+3*sizeof(int)-restsize, restsize /* byte */, &transferred /* byte */)) {
         fprintf(stderr,"function out() failed\n"); 
         goto fin; 
            }
//      fprintf(stdout,"0x%08lx of 0x%08x bytes written (including 12 byte for command words)\n",
//              transferred,bufsize+3*sizeof(int)); 
//      if(transferred) 
//      display_le(stdout,data+3*sizeof(int),bufsize,0);
      CloseHandle(hFile);
      }
   else if(somestr[0]=='e' || somestr[0]=='E') goto fin;
/***
 * exit command   
***/
   else {
/***
 * unknown command   
***/
      fprintf(stderr,"Command %s not implemented\n",somestr);
      } 
   }
/*******
 * Action ends here
*******/
//fprintf(stdout,"Type return to get out of here...");
//fscanf(stdin,"%1s",stmp);
//fscanf(stdin,"%*[^\n]");

fin:

free(functionClassDeviceData);


//  Cleanup
SetupDiDestroyDeviceInfoList(hDevInfo);
#ifdef NOTNOW
#endif /* NOTNOW */
return 0;
}
